%{

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "ssfilter.h"

typedef struct ssfilter * ssfilter_t;

#define YYSTYPE ssfilter_t

static struct ssfilter * alloc_node(int type, void *pred)
{
	struct ssfilter *n = malloc(sizeof(*n));
	if (n == NULL)
		abort();
	n->type = type;
	n->pred = pred;
	n->post = NULL;
	return n;
}

static char		**yy_argv;
static int		yy_argc;
static ssfilter_t	*yy_ret;

#define YYERROR_VERBOSE 1

static int yylex(void);

static void yyerror(char *s)
{
	fprintf(stderr, "ss: bison bellows (while parsing filter): \"%s!\"", s);
}

%}

%token HOSTCOND DCOND SCOND DPORT SPORT LEQ GEQ NEQ AUTOBOUND
%left '|'
%left '&'
%nonassoc '!'

%%
applet: null expr
        {
                *yy_ret = $2;
                $$ = $2;
        }
        | null
        ;
null:   /* NOTHING */ { $$ = NULL; }
        ;
expr:	DCOND HOSTCOND
        {
		$$ = alloc_node(SSF_DCOND, $2);
        }
        | SCOND HOSTCOND
        { 
		$$ = alloc_node(SSF_SCOND, $2);
        }
        | DPORT GEQ HOSTCOND
        {
                $$ = alloc_node(SSF_D_GE, $3);
        }
        | DPORT LEQ HOSTCOND
        {
                $$ = alloc_node(SSF_D_LE, $3);
        }
        | DPORT '>' HOSTCOND
        {
                $$ = alloc_node(SSF_NOT, alloc_node(SSF_D_LE, $3));
        }
        | DPORT '<' HOSTCOND
        {
                $$ = alloc_node(SSF_NOT, alloc_node(SSF_D_GE, $3));
        }
        | DPORT '=' HOSTCOND
        {
		$$ = alloc_node(SSF_DCOND, $3);
        }
        | DPORT NEQ HOSTCOND
        {
		$$ = alloc_node(SSF_NOT, alloc_node(SSF_DCOND, $3));
        }

        | SPORT GEQ HOSTCOND
        {
                $$ = alloc_node(SSF_S_GE, $3);
        }
        | SPORT LEQ HOSTCOND
        {
                $$ = alloc_node(SSF_S_LE, $3);
        }
        | SPORT '>' HOSTCOND
        {
                $$ = alloc_node(SSF_NOT, alloc_node(SSF_S_LE, $3));
        }
        | SPORT '<' HOSTCOND
        {
                $$ = alloc_node(SSF_NOT, alloc_node(SSF_S_GE, $3));
        }
        | SPORT '=' HOSTCOND
        {
		$$ = alloc_node(SSF_SCOND, $3);
        }
        | SPORT NEQ HOSTCOND
        {
		$$ = alloc_node(SSF_NOT, alloc_node(SSF_SCOND, $3));
        }

        | AUTOBOUND
        {
                $$ = alloc_node(SSF_S_AUTO, NULL);
        }
        | expr '|' expr
        {
                $$ = alloc_node(SSF_OR, $1);
	        $$->post = $3;
        }
        | expr expr
        {
                $$ = alloc_node(SSF_AND, $1);
	        $$->post = $2;
        }
        | expr '&' expr

        {
                $$ = alloc_node(SSF_AND, $1);
	        $$->post = $3;
        }
        | '!' expr
        {
                $$ = alloc_node(SSF_NOT, $2);
        }
        | '(' expr ')'
        {
                $$ = $2;
        }
;
%%

int yylex(void)
{
	static int argc;

	if (argc >= yy_argc)
		return 0;

	argc++;
	if (strcmp(yy_argv[argc-1], "!") == 0 ||
	    strcmp(yy_argv[argc-1], "not") == 0)
		return '!';
	if (strcmp(yy_argv[argc-1], "&") == 0 ||
	    strcmp(yy_argv[argc-1], "&&") == 0 ||
	    strcmp(yy_argv[argc-1], "and") == 0)
		return '&';
	if (strcmp(yy_argv[argc-1], "|") == 0 ||
	    strcmp(yy_argv[argc-1], "||") == 0 ||
	    strcmp(yy_argv[argc-1], "or") == 0)
		return '|';
	if (strcmp(yy_argv[argc-1], "(") == 0)
		return '(';
	if (strcmp(yy_argv[argc-1], ")") == 0)
		return ')';
	if (strcmp(yy_argv[argc-1], "dst") == 0)
		return DCOND;
	if (strcmp(yy_argv[argc-1], "src") == 0)
		return SCOND;
	if (strcmp(yy_argv[argc-1], "dport") == 0)
		return DPORT;
	if (strcmp(yy_argv[argc-1], "sport") == 0)
		return SPORT;
	if (strcmp(yy_argv[argc-1], ">=") == 0 ||
	    strcmp(yy_argv[argc-1], "geq") == 0)
		return GEQ;
	if (strcmp(yy_argv[argc-1], "<=") == 0 ||
	    strcmp(yy_argv[argc-1], "leq") == 0)
		return LEQ;
	if (strcmp(yy_argv[argc-1], "!=") == 0 ||
	    strcmp(yy_argv[argc-1], "neq") == 0)
		return NEQ;
	if (strcmp(yy_argv[argc-1], "=") == 0 ||
	    strcmp(yy_argv[argc-1], "==") == 0 ||
	    strcmp(yy_argv[argc-1], "eq") == 0)
		return '=';
	if (strcmp(yy_argv[argc-1], ">") == 0 ||
	    strcmp(yy_argv[argc-1], "gt") == 0)
		return '>';
	if (strcmp(yy_argv[argc-1], "<") == 0 ||
	    strcmp(yy_argv[argc-1], "lt") == 0)
		return '<';
	if (strcmp(yy_argv[argc-1], "autobound") == 0)
		return AUTOBOUND;
	yylval = (void*)parse_hostcond(yy_argv[argc-1]);
	if (yylval == NULL) {
		fprintf(stderr, "Cannot parse dst/src address.\n");
		exit(1);
	}
	return HOSTCOND;
}

int ssfilter_parse(struct ssfilter **f, int argc, char **argv)
{
	yy_argc = argc;
	yy_argv = argv;
	yy_ret  = f;

	if (yyparse()) {
		fprintf(stderr, " Sorry.\n");
		return -1;
	}
	return 0;
}
